import * as fs from 'node:fs';
import { STRUCTURE_FILES, EXTENSION, formatFile } from './common';
import { getStructure } from './map';
import { Struct } from './struct';

// Arguments
if (process.argv.length < 5) {
    console.log('USAGE: npm start -- <Source Output File> <Header Output File> <Input Files...>');
    process.exit(1);
}
process.argv.shift();
process.argv.shift();
function invalidFileType(file: string) {
    throw new Error(`Invalid File Type: ${file}`);
}
const sourceOutput = process.argv.shift()!;
if (!sourceOutput.endsWith('.cpp')) {
    invalidFileType(sourceOutput);
}
const headerOutput = process.argv.shift()!;
if (!headerOutput.endsWith('.h')) {
    invalidFileType(headerOutput);
}
const extraHeaderFiles: string[] = [];
while (process.argv.length > 0) {
    const file = process.argv.shift()!;
    if (file.endsWith(EXTENSION)) {
        // Get Class Name
        const parts = file.split('/');
        let name = parts[parts.length - 1]!;
        name = name.substring(0, name.length - EXTENSION.length);
        // Store
        if (name in STRUCTURE_FILES) {
            throw new Error(`Multiple Definition Files Provided: ${name}${EXTENSION}`);
        }
        STRUCTURE_FILES[name] = file;
    } else if (file.endsWith('.h')) {
        // Extra Headers
        extraHeaderFiles.push(file);
    } else {
        // Invalid File Type
        invalidFileType(file);
    }
}

// Load And Sort Structures
function loadSymbols() {
    // Load
    const structureObjects = [];
    for (const name in STRUCTURE_FILES) {
        structureObjects.push(getStructure(name));
    }
    // Sort
    structureObjects.sort((a, b) => {
        if (a.getName() > b.getName()) {
            return 1;
        } else if (a.getName() < b.getName()) {
            return -1;
        } else {
            return 0;
        }
    });
    // Return
    return structureObjects;
}

// Sort Structures By Dependency
function dependencySort(structureObjects: Struct[]) {
    let loops = 0;
    const MAX_LOOPS = 100;
    while (true) {
        if (loops > MAX_LOOPS) {
            throw new Error('Unable To Sort Dependencies');
        }
        loops++;
        let valid = true;
        // Loop Through Structures
        for (const structure of structureObjects) {
            // Loop Through Dependencies
            const currentIndex = structureObjects.indexOf(structure);
            for (const dependency of structure.getDependencies()) {
                // Compare Current And Dependency Index
                const obj = getStructure(dependency);
                const dependencyIndex = structureObjects.indexOf(obj);
                if (dependencyIndex > currentIndex) {
                    // Dependency Must Be Moved
                    structureObjects.splice(dependencyIndex, 1);
                    structureObjects.splice(currentIndex, 0, obj);
                    valid = false;
                    break;
                }
            }
            // Check If Any Changes Were Made
            if (!valid) {
                break;
            }
        }
        // Check If Any Changes Were Made
        if (valid) {
            break;
        }
    }
}

// Generate Part Of Header
function makeHeaderPart() {
    // Load Symbols
    const structureObjects = loadSymbols();

    // Sort Structures By Dependency
    dependencySort(structureObjects);

    // Generate Code
    let structures = '';
    for (const structure of structureObjects) {
        const name = structure.getName();
        structures += `// ${name}\n`;
        try {
            structures += structure.generate();
        } catch (e) {
            console.log(`Error Generating Header: ${name}: ${e instanceof Error ? e.stack : e}`);
            process.exit(1);
        }
        structures += '\n';
    }

    // Return
    return structures;
}

// Create Main Header
function makeMainHeader(output: string) {
    // Forward Declarations
    let forwardDeclarations = '';
    for (const name in STRUCTURE_FILES) {
        forwardDeclarations += `typedef struct ${name} ${name};\n`;
    }
    forwardDeclarations = forwardDeclarations.trim();
    // Extra Headers
    let extraHeaders = '';
    for (const file of extraHeaderFiles) {
        extraHeaders += fs.readFileSync(file, {encoding: 'utf8'});
    }
    extraHeaders = extraHeaders.trim();
    // Main
    const main = makeHeaderPart().trim();
    // Write
    const result = formatFile('out.h', {forwardDeclarations, extraHeaders, main});
    fs.writeFileSync(output, result);
}
makeMainHeader(headerOutput);

// Generate Compiled Code
function makeCompiledCode(output: string) {
    // Load Symbols
    const structureObjects = loadSymbols();

    // Generate
    let declarations = '';
    for (const structure of structureObjects) {
        const name = structure.getName();
        declarations += `// ${name}\n`;
        try {
            const code = structure.generateCode();
            declarations += code;
        } catch (e) {
            console.log(`Error Generating Code: ${name}: ${e instanceof Error ? e.stack : e}`);
            process.exit(1);
        }
        declarations += '\n';
    }

    // Write
    const headerPath = fs.realpathSync(headerOutput);
    const main = declarations.trim();
    const result = formatFile('out.cpp', {headerPath, main});
    fs.writeFileSync(output, result);
}
makeCompiledCode(sourceOutput);
