import { INDENT, POINTER_SIZE, assertSize, getSelfArg, preventConstruction, toHex } from './common';
import { Method } from './method';
import { Property } from './property';

const structuresWithVTableAddress: string[] = [];
export class VTable {
    readonly #self: string;
    #address: number | null;
    #size: number | null;
    readonly #methods: Method[];
    readonly property: Property;
    #destructorOffset: number;

    // Constructor
    constructor(self: string) {
        this.#self = self;
        this.#address = null;
        this.#size = null;
        this.#methods = [];
        this.#destructorOffset = 0;
        // Create Property
        this.property = new Property(0, this.#getName() + ' *', 'vtable', this.#self);
    }

    // Setters
    setAddress(address: number) {
        this.#address = address;
        structuresWithVTableAddress.push(this.#self);
    }
    setSize(size: number) {
        this.#size = size;
    }
    setDestructorOffset(destructorOffset: number) {
        this.#destructorOffset = destructorOffset;
    }

    // Add To VTable
    #add(target: Method[], method: Method) {
        // Check Offset
        const offset = method.address;
        if ((offset % POINTER_SIZE) !== 0) {
            throw new Error(`Invalid VTable Offset: ${toHex(offset)}`);
        }
        // Add
        const index = offset / POINTER_SIZE;
        if (target[index]) {
            throw new Error(`Duplicate Virtual Method At Offset: ${toHex(offset)}`);
        }
        target[index] = method;
    }
    add(method: Method) {
        this.#add(this.#methods, method);
    }

    // Get Structure Name
    #getName() {
        return this.#self + '_vtable';
    }

    // Check
    #check() {
        // Check Size
        if (this.#size !== null) {
            const size = this.#size;
            const maxMethodCount = size / POINTER_SIZE;
            if (maxMethodCount < this.#methods.length) {
                throw new Error(`VTable Size Too Small: ${toHex(size)}`);
            }
            if (maxMethodCount > this.#methods.length) {
                this.#methods.length = maxMethodCount;
            }
        }
    }

    // Get Full Methods Table
    getMethods() {
        // Copy Array
        const out = [];
        out.push(...this.#methods);
        // Add Destructors (https://stackoverflow.com/a/17960941)
        const destructor_return = `${this.#self} *`;
        const destructor_args = `(${getSelfArg(this.#self)})`;
        this.#add(out, new Method(this.#self, 'destructor_complete', destructor_return, destructor_args, 0x0 + this.#destructorOffset, false, false));
        this.#add(out, new Method(this.#self, 'destructor_deleting', destructor_return, destructor_args, 0x4 + this.#destructorOffset, false, false));
        // Return
        return out;
    }

    // Get Parent Method
    #getParentSelf(method: Method, parent: string | null) {
        if (method.isInherited) {
            // Parent Exists
            let out: string;
            if (structuresWithVTableAddress.includes(parent!)) {
                out = parent!;
            } else {
                // Unable To Determine
                out = this.#self;
            }
            return out;
        } else {
            // Method Has No Parent
            return undefined;
        }
    }

    // Generate Header Code
    generate() {
        let out = '';

        // Check
        this.#check();

        // Wrappers
        const methods = this.getMethods();
        for (const info of methods) {
            if (info) {
                out += info.generate(false, true);
            }
        }

        // Structure
        out += `typedef struct ${this.#getName()} ${this.#getName()};\n`;
        out += `struct ${this.#getName()} {\n`;
        for (let i = 0; i < methods.length; i++) {
            const info = methods[i];
            if (info) {
                out += info.getProperty();
            } else {
                out += `${INDENT}void *unknown${i};\n`;
            }
        }
        if (this.#size === null) {
            // Prevent Construction
            out += preventConstruction(this.#getName());
        }
        out += `};\n`;

        // Sanity Check Size
        if (this.#size !== null) {
            out += assertSize(this.#getName(), this.#size);
        }

        // Pointers
        if (this.#address !== null) {
            // Base
            out += `extern ${this.#getName()} *${this.#getName()}_base;\n`;
        }

        // Return
        return out;
    }

    // Generate Compiled Code
    generateCode(directParent: string | null) {
        let out = '';

        // Check
        this.#check();

        // Pointers
        if (this.#address !== null) {
            // Base
            out += `${this.#getName()} *${this.#getName()}_base = (${this.#getName()} *) ${toHex(this.#address)};\n`;
            // Methods
            const methods = this.getMethods();
            for (const info of methods) {
                if (info) {
                    out += info.generate(true, true, this.#getParentSelf(info, directParent));
                }
            }
        }

        // Return
        return out;
    }
}